@@ -387,7 +387,7 @@ def numeric_convert_digits(d):
 			yield int(x)
 
 numeric_signs = {
-	16384 : 1,
+	numeric_negative : 1,
 }
 
 def numeric_unpack(x):
@@ -421,6 +421,9 @@ oid_to_io = {
 	pg_types.CIRCLEOID : (circle_pack, circle_unpack),
 }
 
+oid_to_io[pg_types.CIDROID] = (None, None)
+oid_to_io[pg_types.INETOID] = (None, None)
+
 def process_tuple(procs, tup, exception_handler):
 	"""
 	Call each item in `procs` with the corresponding
@@ -433,8 +436,10 @@ def process_tuple(procs, tup, exception_handler):
 	"""
 	i = len(procs)
 	if len(tup) != i:
-		raise ValueError(
-			"inconsistent items, %d processors and %d objects"
+		raise TypeError(
+			"inconsistent items, %d processors and %d items in row" %(
+				i, len(tup)
+			)
 		)
 	r = [None] * i
 	try:
@@ -449,8 +454,14 @@ def process_tuple(procs, tup, exception_handler):
 		raise RuntimeError("process_tuple exception handler failed to raise")
 	return r
 
+def process_chunk(procs, tupc, fail):
+	return [
+		process_tuple(procs, x, fail) for x in tupc
+	]
+
 try:
-	from .optimized import process_tuple
+	# C implementation of the tuple processors.
+	from .optimized import process_tuple, process_chunk
 except ImportError:
 	pass
 
@@ -544,7 +555,7 @@ def composite_typio(
 		if len(data) > 80:
 			# Be sure not to fill screen with noise.
 			data = data[:75] + ' ...'
-		te = pg_exc.ColumnError(
+		raise pg_exc.ColumnError(
 			"failed to pack attribute %d, %s::%s, of composite %s for transfer" %(
 				itemnum,
 				attnames[itemnum],
@@ -552,18 +563,17 @@ def composite_typio(
 				composite_name,
 			),
 			details = {
-				'data': data,
+				'context': data,
+				'position' : str(itemnum)
 			},
 		)
-		te.index = itemnum
-		te.raise_exception()
 
 	def raise_unpack_tuple_error(procs, tup, itemnum):
 		data = repr(tup[itemnum])
 		if len(data) > 80:
 			# Be sure not to fill screen with noise.
 			data = data[:75] + ' ...'
-		te = pg_exc.ColumnError(
+		raise pg_exc.ColumnError(
 			"failed to unpack attribute %d, %s::%s, of composite %s from wire data" %(
 				itemnum,
 				attnames[itemnum],
@@ -571,11 +581,10 @@ def composite_typio(
 				composite_name,
 			),
 			details = {
-				'data': data,
+				'context': data,
+				'position' : str(itemnum),
 			},
 		)
-		te.index = itemnum
-		te.raise_exception()
 
 	def unpack_a_record(data):
 		data = tuple([x[1] for x in ts.record_unpack(data)])
@@ -714,8 +723,6 @@ class TypeIO(object, metaclass = ABCMeta):
 			pg_types.VARCHAROID : (None, None),
 			pg_types.CSTRINGOID : (None, None),
 			pg_types.TEXTOID : (None, None),
-			pg_types.CIDROID : (None, None),
-			pg_types.INETOID : (None, None),
 			pg_types.REGPROCEDUREOID : (None, None),
 			pg_types.REGTYPEOID : (None, None),
 			pg_types.REGPROCOID : (None, None),
