@@ -5,7 +5,7 @@
 'PQ version 3.0 client transactions'
 import sys
 import os
-from abc import abstractmethod
+from abc import ABCMeta, abstractmethod
 from pprint import pformat
 from itertools import chain
 from operator import itemgetter
@@ -13,14 +13,12 @@ get0 = itemgetter(0)
 get1 = itemgetter(1)
 
 from ..python.functools import Composition as compose
-from .. import api as pg_api
 from .. import exceptions as pg_exc
 from . import element3 as element
 
 from hashlib import md5
 from ..resolved.crypt import crypt
 
-proctup = compose((get1, compose))
 Receiving = True
 Sending = False
 Complete = (None, None)
@@ -34,26 +32,46 @@ AsynchronousMap = {
 def return_arg(x):
 	return x
 
-class Transaction(pg_api.InterfaceElement):
-	ife_label = 'PROTOCOL'
-	ife_ancestor = None
-	_ife_exclude_snapshot = True
+class Transaction(object, metaclass = ABCMeta):
+	"""
+	If the fatal attribute is not None, an error occurred, and the
+	`error_message` attribute should be set to a element3.Error instance.
+	"""
+	fatal = None
 
 	@abstractmethod
 	def messages_received(self):
 		"""
-		Return an iterable to the messages received.
+		Return an iterable to the messages received that have been processed.
 		"""
 
-	def ife_snapshot_text(self):
-		s = os.linesep
-		s += repr(self)
-		s += os.linesep*2 + '   [Received]' + os.linesep*2
-		s += pformat(list(self.messages_received()))
-		if hasattr(self, 'error_message'):
-			s += os.linesep*2 + '   [Error Message]' + os.linesep*2
-			s += repr(self.error_message)
-		return s
+class Closing(Transaction):
+	"""
+	Send the disconnect message and mark the connection as closed.
+	"""
+	error_message = element.ClientError(
+		# pg_exc.ConnectionDoesNotExistError.code
+		code = '08003',
+		severity = 'FATAL',
+		message = 'operation on closed connection',
+		hint = "A new connection needs to be "\
+			"created in order to query the server.",
+	)
+
+	def messages_received(self):
+		return ()
+
+	def sent(self):
+		"""
+		Empty messages and mark complete.
+		"""
+		self.messages = ()
+		self.fatal = True
+		self.state = Complete
+
+	def __init__(self):
+		self.messages = (element.DisconnectMessage,)
+		self.state = (Sending, self.sent)
 
 class Negotiation(Transaction):
 	"""
@@ -68,8 +86,6 @@ class Negotiation(Transaction):
 	state.
 	"""
 	state = None
-	asynchook = AsynchronousMap
-	fatal = None
 
 	def __init__(self,
 		startup_message : "startup message to send",
@@ -77,8 +93,8 @@ class Negotiation(Transaction):
 	):
 		self.startup_message = startup_message
 		self.password = password
-		self._asyncs = []
 		self.received = [()]
+		self.asyncs = []
 		self.authtype = None
 		self.killinfo = None
 		self.authok = None
@@ -92,23 +108,8 @@ class Negotiation(Transaction):
 		s += pformat((self.startup_message, self.password)).lstrip()
 		return s
 
-	def reset(self):
-		self.authtype = None
-		self.killinfo = None
-		self.last_ready = None
-		self.authok = None
-		self._asyncs.clear()
-		self.received.clear()
-		self.machine = self.state_machine()
-		self.messages = next(self.machine)
-		self.state = (Sending, self.sent)
-
-	def asyncs(self):
-		"iterate over asynchronous messages received"
-		return self._asyncs
-
 	def messages_received(self):
-		return self.received
+		return self.processed
 
 	def sent(self):
 		"""
@@ -126,37 +127,23 @@ class Negotiation(Transaction):
 		if messages is not self.received[-1]:
 			self.received.append(messages)
 		else:
-			# looks like an interrupt occurred, redo everything.
-			self._asyncs.clear()
-			self.machine = self.state_machine()
-			out_messages = next(self.machine)
-			redo_messages = []
-			for xs in self.received:
-				for x in xs:
-					if x[0] == element.Error.type:
-						if self.fatal is None:
-							self.error_message = element.Error.parse(x[1])
-							self.fatal = True
-							self.state = Complete
-							return 0
-					elif x[0] in self.asynchook:
-						self._asyncs.append(self.asynchook[x[0]](x[1]))
-					else:
-						out_messages = self.machine.send(x)
+			raise RuntimeError("negotiation was interrupted")
 
 		# if an Error message was found, complete and leave.
 		count = 0
 		try:
 			for x in messages:
 				count += 1
-				if x[0] == element.Error.type:
+				if x[0] is element.Error.type:
 					if self.fatal is None:
 						self.error_message = element.Error.parse(x[1])
 						self.fatal = True
 						self.state = Complete
-						return 0
-				elif x[0] in self.asynchook:
-					self._asyncs.append(self.asynchook[x[0]](x[1]))
+						return count
+				elif x[0] in AsynchronousMap:
+					self.asyncs.append(
+						AsynchronousMap[x[0]](x[1])
+					)
 				else:
 					out_messages = self.machine.send(x)
 					if out_messages:
@@ -165,26 +152,46 @@ class Negotiation(Transaction):
 			# generator is complete, negotiation is complete..
 			self.state = Complete
 			return count
+
 		if out_messages:
 			self.messages = out_messages
 			self.state = (Sending, self.sent)
 		return count
 
+	def unsupported_auth_request(self, req):
+		self.fatal = True
+		self.error_message = element.ClientError(
+			"unsupported authentication request %r(%d)" %(
+				element.AuthNameMap.get(req, '<unknown>'), req,
+			),
+			code = "--AUT",
+			hint = \
+				"'postgresql.protocol' only supports: " \
+				"MD5, crypt, plaintext, and trust."
+		)
+		self.state = Complete
+
 	def state_machine(self):
 		"""
 		Generator keeping the state of the connection negotiation process.
 		"""
 		x = (yield (self.startup_message,))
 
-		if x[0] != element.Authentication.type:
-			raise pg_exc.ProtocolError(
-				"received message of type {mt}, but expected {et}".format(
-					repr(x[0]),
-					element.Authentication.type
-				)
+		if x[0] is not element.Authentication.type:
+			self.fatal = True
+			self.error_message = element.ClientError(
+				message = \
+					"received message of type %s, " \
+					"but expected %s" %(
+						repr(x[0]),
+						repr(element.Authentication.type)
+					),
+				code = '08P01',
 			)
+			return
 
 		self.authtype = element.Authentication.parse(x[1])
+
 		req = self.authtype.request
 		if req != element.AuthRequest_OK:
 			if req == element.AuthRequest_Cleartext:
@@ -197,59 +204,60 @@ class Negotiation(Transaction):
 			else:
 				##
 				# Not going to work. Sorry :(
-				# The many authentication types supported by PostgreSQL are not easy
-				# to implement, especially when implementations for the type don't exist
-				# for Python.
-				# Worse yet, some of them may want control of the wire..
-				raise pg_exc.AuthenticationMethodError(
-						"unsupported authentication request %r(%d)" %(
-						element.AuthNameMap.get(req, '<unknown>'), req,
-					),
-					details = {
-						'hint' : \
-							"'postgresql.protocol' supports: MD5, crypt, plaintext, and trust."
-					}
-				)
+				# The many authentication types supported by PostgreSQL are not
+				# easy to implement, especially when implementations for the
+				# type don't exist for Python.
+				self.unsupported_auth_request(req)
+				return
 			x = (yield (element.Password(pw),))
-			if x[0] != element.Authentication.type:
-				raise pg_exc.ProtocolError(
-					"received message of type {1}, but expected {2}".format(
-						repr(x[0]),
-						element.Authentication.type
-					)
-				)
-			self.authok = element.Authentication.parse(x[1])
 
+			self.authok = element.Authentication.parse(x[1])
 			if self.authok.request != element.AuthRequest_OK:
-				raise pg_exc.ProtocolError(
-					"expected an OK from the authentication " \
-					"message, but received {1}({2}) instead".format(
-						element.AuthNameMap.get(self.authok.request, '<unknown>'),
-						str(self.authok.request),
-					)
+				self.fatal = True
+				self.error_message = element.ClientError(
+					message = \
+						"expected an OK from the authentication " \
+						"message, but received %s(%s) instead" %(
+							repr(element.AuthNameMap.get(
+								self.authok.request, '<unknown>'
+							)),
+							repr(self.authok.request),
+						),
+					code = "08P01"
 				)
+				return
 		else:	
 			self.authok = self.authtype
 
 		# Done authenticating, pick up the killinfo and the ready message.
 		x = (yield None)
-		if x[0] != element.KillInformation.type:
-			raise pg_exc.ProtocolError(
-				"received message of type {1}, but expected {2}".format(
-					repr(x[0]),
-					element.KillInformation.type
-				)
+		if x[0] is not element.KillInformation.type:
+			self.fatal = True
+			self.error_message = element.ClientError(
+				message = \
+					"received message of type %s, " \
+					"but expected %s" %(
+						repr(x[0]),
+						element.KillInformation.type
+					),
+				code = "08P01"
 			)
+			return
 		self.killinfo = element.KillInformation.parse(x[1])
 
 		x = (yield None)
-		if x[0] != element.Ready.type:
-			raise pg_exc.ProtocolError(
-				"received message of type {1}, but expected {2}".format(
-					repr(x[0]),
-					element.Ready.type
-				)
+		if x[0] is not element.Ready.type:
+			self.fatal = True
+			self.error_message = element.ClientError(
+				message = \
+					"received message of type %s, " \
+					"but expected %s" %(
+						repr(x[0]),
+						repr(element.Ready.type)
+					),
+				code = "08P01"
 			)
+			return
 		self.last_ready = element.Ready.parse(x[1])
 
 class Instruction(Transaction):
@@ -258,22 +266,21 @@ class Instruction(Transaction):
 	It provides the messages to be sent and takes the response messages for order
 	and integrity validation:
 
-		Instruction([postgresql.protocol.element3.Message(), ..])
+		Instruction([.element3.Message(), ..])
 
 	A message must be one of:
 
-		* `postgresql.protocol.element3.Query`
-		* `postgresql.protocol.element3.Function`
-		* `postgresql.protocol.element3.Parse`
-		* `postgresql.protocol.element3.Bind`
-		* `postgresql.protocol.element3.Describe`
-		* `postgresql.protocol.element3.Close`
-		* `postgresql.protocol.element3.Execute`
-		* `postgresql.protocol.element3.Synchronize`
-		* `postgresql.protocol.element3.Flush`
+		* `.element3.Query`
+		* `.element3.Function`
+		* `.element3.Parse`
+		* `.element3.Bind`
+		* `.element3.Describe`
+		* `.element3.Close`
+		* `.element3.Execute`
+		* `.element3.Synchronize`
+		* `.element3.Flush`
 	"""
 	state = None
-	fatal = None
 	CopyFailMessage = element.CopyFail(b"invalid termination")
 
 	# The hook is the dictionary that provides the path for the
@@ -378,40 +385,42 @@ class Instruction(Transaction):
 		element.Flush.type : None,
 	}
 
-	# This map provides parsers for asynchronous messages
-	asynchook = AsynchronousMap
-
 	initial_state = (
 		(),     # last messages,
 		(0, 0), # request position, response position
 		(0, 0), # last request position, last response position
 	)
 
-	def __init__(self, commands):
+	def __init__(self, commands, asynchook = return_arg):
 		"""
 		Initialize an `Instruction` instance using the given commands.
 
 		Commands are `postgresql.protocol.element3.Message` instances:
 
-		 * `postgresql.protocol.element3.Query`
-		 * `postgresql.protocol.element3.Function`
-		 * `postgresql.protocol.element3.Parse`
-		 * `postgresql.protocol.element3.Bind`
-		 * `postgresql.protocol.element3.Describe`
-		 * `postgresql.protocol.element3.Close`
-		 * `postgresql.protocol.element3.Execute`
-		 * `postgresql.protocol.element3.Synchronize`
-		 * `postgresql.protocol.element3.Flush`
+		 * `.element3.Query`
+		 * `.element3.Function`
+		 * `.element3.Parse`
+		 * `.element3.Bind`
+		 * `.element3.Describe`
+		 * `.element3.Close`
+		 * `.element3.Execute`
+		 * `.element3.Synchronize`
+		 * `.element3.Flush`
 		"""
 		# Commands are accessed by index.
 		self.commands = tuple(commands)
+		self.asynchook = asynchook
+		self.completed = []
+		self.last = self.initial_state
+		self.messages = list(self.commands)
+		self.state = (Sending, self.standard_sent)
+		self.fatal = None
 
 		for cmd in self.commands:
 			if cmd.type not in self.hook:
 				raise TypeError(
 					"unknown message type for PQ 3.0 protocol", cmd.type
 				)
-		self.reset()
 
 	def __repr__(self):
 		return '%s.%s(%s%s)' %(
@@ -421,26 +430,6 @@ class Instruction(Transaction):
 			pformat(self.commands)
 		)
 
-	def reset(self):
-		"""
-		Reset the `Transaction` instance to its initial state.
-		"""
-		self.completed = []
-		self._asyncs = []
-		self.last = self.initial_state
-		self.messages = self.commands
-		if self.messages:
-			self.state = (Sending, self.standard_sent)
-		self.fatal = None
-		if hasattr(self, 'error_message'):
-			del self.error_message
-		if hasattr(self, 'last_ready'):
-			del self.last_ready
-
-	def asyncs(self):
-		"iterate over asynchronous messages received"
-		return chain.from_iterable(map(get1, self._asyncs))
-
 	def messages_received(self):
 		'Received and validate messages'
 		return chain.from_iterable(map(get1, self.completed))
@@ -469,12 +458,14 @@ class Instruction(Transaction):
 		# so go ahead and reprocess it.
 		if messages is self.last[0]:
 			offset, current_step = self.last[1]
+			# don't clear the asyncs. they have already been process by the hook.
 		else:
 			offset, current_step = self.last[2]
+			# it's a new set, so we can clear the asyncs record.
+			self._asyncs = []
 		cmd = self.commands[offset]
 		paths = self.hook[cmd.type]
 		processed = []
-		asyncs = []
 		count = 0
 
 		for x in messages:
@@ -485,12 +476,15 @@ class Instruction(Transaction):
 
 			if path is None:
 				# No path for message type, could be a protocol error.
-				if x[0] == element.Error.type:
+				if x[0] is element.Error.type:
 					em = element.Error.parse(x[1])
 					fatal = em['severity'].upper() in (b'FATAL', b'PANIC')
-					if fatal or not hasattr(self, 'error_message'):
-						self.error_message = em
-						self.fatal = fatal
+					self.error_message = em
+					self.fatal = fatal
+					if fatal is True:
+						# can't sync up if it's fatal.
+						self.state = Complete
+						return count
 					# Error occurred, so sync up with backend if
 					# the current command is not 'Q' or 'F' as they
 					# imply a sync message.
@@ -513,23 +507,31 @@ class Instruction(Transaction):
 					# On a new command, setup the new step.
 					current_step = 0
 					continue
-				elif x[0] in self.asynchook:
-					asyncs.append(self.asynchook[x[0]](x[1]))
+				elif x[0] in AsynchronousMap:
+					if x not in self._asyncs:
+						msg = AsynchronousMap[x[0]](x[1])
+						try:
+							self.asynchook(msg)
+						except Exception as err:
+							# exception thrown by async message handler?
+							# notify the user, but continue...
+							sys.excepthook(*sys.exc_info())
+						# it's been processed, so don't process it again.
+						self._asyncs.append(x)
 				else:
 					##
-					# Procotol violation
-					err = pg_exc.ProtocolError(
+					# Procotol violation.
+					self.error_message = element.ClientError(
 						"expected message of types %r, " \
 						"but received %r instead" % (
 							tuple(paths[current_step].keys()), x[0]
 						),
-						source = 'DRIVER',
-						details = {
-							'severity': 'FATAL',
-						}
+						code = '08P01',
+						severity = 'FATAL',
 					)
-					self.ife_descend(err)
-					err.raise_exception()
+					self.fatal = True
+					self.state = Complete
+					return count
 			else:
 				# Valid message
 				r = path(x[1])
@@ -566,12 +568,6 @@ class Instruction(Transaction):
 		# have not been put there already.
 		if not self.completed or self.completed[-1][0] != id(messages):
 			self.completed.append((id(messages), processed))
-		if not self._asyncs or self._asyncs[-1][0] != id(messages):
-			self._asyncs.append((id(messages), asyncs or ()))
-			# Remove empty async items before the last
-			if len(self._asyncs) > 1:
-				if not self._asyncs[-2][1]:
-					del self._asyncs[-2]
 
 		# Store the state for the next transition.
 		self.last = (messages, self.last[2], (offset, current_step),)
@@ -586,11 +582,11 @@ class Instruction(Transaction):
 			last = processed[-1]
 			if type(last) is bytes:
 				self.state = (Receiving, self.put_copydata)
-			elif last.type == element.CopyToBegin.type:
+			elif last.type is element.CopyToBegin.type:
 				self.state = (Receiving, self.put_copydata)
 			elif last.type is element.Tuple.type:
 				self.state = (Receiving, self.put_tupledata)
-			elif last.type == element.CopyFromBegin.type:
+			elif last.type is element.CopyFromBegin.type:
 				self.CopyFailSequence = (self.CopyFailMessage,) + \
 					self.commands[offset+1:]
 				self.CopyDoneSequence = (element.CopyDoneMessage,) + \
@@ -603,25 +599,18 @@ class Instruction(Transaction):
 		In the context of a copy, `put_copydata` is used as a fast path for
 		storing `element.CopyData` messages. When a non-`element.CopyData.type`
 		message is received, it reverts the ``state`` attribute back to
-		`standard_put` to process the message..
+		`standard_put` to process the message-sequence.
 		"""
 		# "Fail" quickly if the last message is not copy data.
 		if messages[-1][0] is not element.CopyData.type:
 			self.state = (Receiving, self.standard_put)
 			return self.standard_put(messages)
 
-		# Build a sequence of raw copy data.
-		lines = []
-		for x in messages:
-			# While this case should probably never happen as protection
-			# against this it for valid clients is employed directly above,
-			# it's important that we switch back to `standard_put`
-			# if something other than `CopyData` is received as `put_copydata`
-			# cannot handle anything other than `CopyData`
-			if x[0] is not element.CopyData.type:
-				self.state = (Receiving, self.standard_put)
-				return self.standard_put(messages)
-			lines.append(x[1])
+		cdt = element.CopyData.type
+		lines = [x[1] for x in messages if x[0] is cdt]
+		if len(lines) != len(messages):
+			self.state = (Receiving, self.standard_put)
+			return self.standard_put(messages)
 
 		if not self.completed or self.completed[-1][0] != id(messages):
 			self.completed.append((id(messages), lines))
@@ -640,13 +629,11 @@ class Instruction(Transaction):
 			return self.standard_put(messages)
 
 		p = element.Tuple.parse
-		tuplemessages = []
-		for x in messages:
-			if x[0] is not element.Tuple.type:
-				self.state = (Receiving, self.standard_put)
-				return self.standard_put(messages)
-			tuplemessages.append(p(x[1]))
-		#tuplemessages = list(map(proctup, messages))
+		t = element.Tuple.type
+		tuplemessages = [p(x[1]) for x in messages if x[0] is t]
+		if len(tuplemessages) != len(messages):
+			self.state = (Receiving, self.standard_put)
+			return self.standard_put(messages)
 
 		if not self.completed or self.completed[-1][0] != id(messages):
 			self.completed.append(((id(messages), tuplemessages)))
@@ -668,13 +655,13 @@ class Instruction(Transaction):
 		"""
 		The state method for sending copy data.
 
-		After each call to `sent_from_stdin`, the `messages` attribute is set to a
-		`CopyFailSequence`. This sequence of messages assures that the COPY will be
-		properly terminated.
+		After each call to `sent_from_stdin`, the `messages` attribute is set
+		to a `CopyFailSequence`. This sequence of messages assures that the
+		COPY will be properly terminated.
 
 		If new copy data is not provided, or `messages` is *not* set to
-		`CopyDoneSequence`, the transaction will instruct the remote end to cause
-		the COPY to fail.
+		`CopyDoneSequence`, the transaction will instruct the remote end to
+		cause the COPY to fail.
 		"""
 		if self.messages is self.CopyDoneSequence or \
 		self.messages is self.CopyFailSequence:
